from threading import Thread
import time
import threading
import configparser
from enum import IntEnum
import argparse
from psc.pgstatlogger.pg_stat_logger import PSCLogger
from psc.pgstatcommon.pg_stat_common import *
import psc.postgresql as postgresql
import logging
import hashlib
from enum import Enum
import itertools

VERSION = 1.2


class SysConf:
    def __init__(self):
        self.current_dir = os.path.dirname(os.path.realpath(__file__))
        self.config = configparser.RawConfigParser()
        self.config.optionxform = lambda option: option
        self.config.read(
            os.path.join(
                self.current_dir,
                'conf',
                os.path.splitext(os.path.basename(__file__))[0] + ".conf"
            )
        )

        def get_key(section, parameter, default, boolean=False):
            try:
                return read_conf_param_value(self.config[section][parameter], boolean)
            except KeyError:
                return default

        self.dbs_dict = {}
        for db in self.config['databases']:
            self.dbs_dict[db] = read_conf_param_value(self.config['databases'][db])

        # main parameters
        self.application_name = get_key('main', 'application_name', 'custom_soft_reindex')
        self.postfix_str = get_key('main', 'postfix_str', '_tmp_csr')
        self.threads_num = int(get_key('main', 'threads_num', '2'))
        self.execute_sql = get_key('main', 'execute_sql', 'True', boolean=True)
        self.lock_observer_sleep_interval = int(
            get_key('main', 'lock_observer_sleep_interval', '5')
        )
        self.conn_exception_sleep_interval = int(
            get_key('main', 'conn_exception_sleep_interval', '5')
        )
        self.a_share_lock_sleep_interval = int(
            get_key('main', 'a_share_lock_sleep_interval', '1')
        )
        self.cancel_blocker_tx_timeout = get_key('main', 'cancel_blocker_tx_timeout', '5 seconds')
        self.application_max_mem = get_key('main', 'application_max_mem', '2000')
        self.ban_hold_iterations = int(
            get_key('main', 'ban_hold_iterations', '10')
        )
        self.max_dependent_objects = int(
            get_key('main', 'max_dependent_objects', '100')
        )

        # log parameters
        self.log_level = get_key('log', 'log_level', 'Info')
        self.log_sql = int(get_key('log', 'log_sql', '1'))
        self.file_maxmbytes = int(get_key('log', 'file_maxmbytes', '50'))
        self.file_backupcount = int(get_key('log', 'file_backupcount', '5'))

        # session parameters
        self.deadlock_timeout = get_key('postgresql', 'deadlock_timeout', '100ms')
        self.statement_timeout = get_key('postgresql', 'statement_timeout', '1h')
        self.work_mem = get_key('postgresql', 'work_mem', '200MB')
        self.maintenance_work_mem = get_key('postgresql', 'maintenance_work_mem', '1GB')


class BasicEnum:
    def __str__(self):
        return self.value


class AtomicStageType(BasicEnum, Enum):
    INDEX_CREATE = 'INDEX_CREATE'
    FK_DROP = 'FK_DROP'
    FK_USING_INDEX_ADD = 'FK_USING_INDEX_ADD'
    FK_NOT_VALID_ADD = 'FK_NOT_VALID_ADD'
    FK_VALIDATE = 'FK_VALIDATE'


class FldsEnumBasicList(IntEnum):   # get indexes query
    relnamespace = 0
    schema_name = 1
    tbl_name = 2
    idx_name = 3
    idx_def = 4
    indexrelid = 5					# index oid
    indrelid = 6					# table oid
    complex = 7


class FldsEnumComplex(IntEnum):     # for each complex case
    schema_name = 0
    tbl_name = 1
    idx_name = 2
    indisprim = 3
    indisunique = 4
    contype = 5
    condef = 6
    conn_schema_name = 7
    conn_tbl_name = 8
    conn_name = 9


class FldsEnumIntegrity(IntEnum):   # validation query
    schema_name = 0
    tbl_name = 1
    idx_name = 2
    indisprim = 3
    indisunique = 4
    idxdef = 5
    contype = 6
    condef = 7
    conn_schema_name = 8
    conn_tbl_name = 9
    conn_name = 10


class CSRGlobal:            # custom soft reindex global vars and objects
    idxs = {}                    # key is indexrelid
    active_idxs = {}             # key is indrelid
    active_rels = {}             # key is indrelid, value is list with linked tables by FK
    atomic_stages = {}           # key is indexrelid, value is list of tuples [(AtomicStageType, md5(DDL), DDL), ...]
    complex_cases = {}    # key is indexrelid, value is result of function open_complex_case
    db_conns = {}
    is_terminate = False
    total_idxs = 0
    processed_idx = 0
    lock = threading.Lock()
    sys_conf = None
    logger = None
    args = None
    workers_db_pid = []
    rules = {}            # array of objects from "reindex_scenarios/rules/<reindex_scenario>.py"
    backup_dir = None
    ban_idxs = []
    file_lock = None

    def __init__(self):
        try:
            parser = argparse.ArgumentParser()
            parser.add_argument(
                "--reindex-scenario",
                help="Select specific SQL file in 'reindex_scenarios' directory",
                type=str
            )
            parser.add_argument(
                "--test",
                help="Run unit test from 'tests' directory using corresponding scenario" +
                     "from 'reindex_scenarios' directory",
                type=str
            )
            parser.add_argument(
                "--test-validate",
                help="Validate test results",
                type=str
            )
            parser.add_argument(
                "--db-name",
                help="Select DB name from [databases] section of 'conf/custom_soft_reindex.conf'",
                type=str,
                default="ALL"
            )
            parser.add_argument(
                "--track-actions",
                help="Write actions to public.csr_actions",
                action='store_true',
                default=False
            )
            parser.add_argument(
                "--backup-dir",
                help="Directory for saving backups",
                type=str,
                default="backup"
            )
            parser.add_argument(
                "--disable-file-lock",
                help="Don't check '.lock' file",
                action='store_true',
                default=False
            )
            parser.add_argument(
                "--skip-fk",
                help="Do not rebuild indexes on which foreign keys depend (complex cases)",
                action='store_true',
                default=False
            )
            parser.add_argument(
                "--perm",
                help="Run in permanent mode with sleep interval in hours",
                type=int,
                default=0
            )
            parser.add_argument(
                "--version",
                help="Show the version number and exit",
                action='store_true',
                default=False
            )
            parser.add_argument(
                "--action-expiration-hours",
                help="Number of hours after which the action is considered expired",
                type=int,
                default=0
            )

            try:
                self.args = parser.parse_args()
            except SystemExit as e:
                sys.exit(0)
            except:
                print(exception_helper())
                sys.exit(0)

            if not len(sys.argv) > 1:
                print("No arguments. Type -h for help.")
                sys.exit(0)

            if self.args.version:
                print("Version %s" % VERSION)
                sys.exit(0)

            if self.args.perm > 0 and self.args.track_actions:
                print("Invalid options: --perm and --track_actions is not compatible!")
                sys.exit(0)

            if self.args.perm > 0 and self.args.test:
                print("Invalid options: --perm and --test is not compatible!")
                sys.exit(0)

            self.sys_conf = SysConf()

            if self.args.test is not None and len(self.args.test) != 0:
                scenario_name = self.args.test
                self.args.reindex_scenario = self.args.test
            else:
                scenario_name = os.path.splitext(self.args.reindex_scenario)[0]

            if self.args.backup_dir != "backup":
                self.backup_dir = self.args.backup_dir + "/" + scenario_name + "/" + self.args.db_name
            else:
                self.backup_dir = os.path.join(self.sys_conf.current_dir, "backup", scenario_name, self.args.db_name)

            prepare_dirs(
                self.sys_conf.current_dir,
                ["log", self.backup_dir]
            )
            self.logger = PSCLogger(
                self.sys_conf.application_name + "_" + self.args.db_name + "_" + scenario_name,
                log_level=logging._nameToLevel[self.sys_conf.log_level.upper()],
                max_bytes=1024 * 1000 * self.sys_conf.file_maxmbytes,
                backup_count=self.sys_conf.file_backupcount,
                delay=3
            )
            self.logger.start()

            if not self.args.disable_file_lock:
                self.file_lock = AppFileLock(            # AppFileLock used for multiple processes
                    self.sys_conf.current_dir,
                    self.sys_conf.application_name + "_" + scenario_name + "_" + self.args.db_name
                )
            limit_memory(1024 * 1000 * int(read_conf_param_value(self.sys_conf.application_max_mem)))
        except SystemExit as e:
            print("Exiting...")
            sys.exit(0)
        except:
            print("Can't initialize application. Exiting...")
            print(exception_helper())
            sys.exit(0)

    def reset(self):
        self.idxs.clear()
        self.active_idxs.clear()
        self.active_rels.clear()
        self.total_idxs = 0
        self.processed_idx = 0
        del self.workers_db_pid[:]
        self.rules.clear()
        self.atomic_stages.clear()
        self.complex_cases.clear()
        self.db_conns.clear()


def lock_observer(thread_name, str_conn, app_name_postfix):
    def sleep_lo():
        for i in range(50):
            time.sleep(CSR.sys_conf.lock_observer_sleep_interval / 50)
            if CSR.is_terminate:
                break

    sleep_lo()
    CSR.logger.log('Thread \'%s\' runned! Observed pids: %s' % (thread_name, str(CSR.workers_db_pid)), "Info")
    do_work = True
    while do_work:
        do_work = False
        try:
            db_conn = postgresql.open(str_conn)
            db_conn.execute("SET application_name = '%s'" % (CSR.sys_conf.application_name + "_" + app_name_postfix))

            while len(CSR.workers_db_pid) > 0:
                for pid in CSR.workers_db_pid:
                    pid_is_locker = get_scalar(db_conn, """
                        SELECT exists(
                            SELECT 1
                            FROM pg_locks waiting
                            JOIN pg_stat_activity waiting_stm ON waiting_stm.pid = waiting.pid
                            JOIN pg_locks other ON waiting.database = other.database
                                AND waiting.relation = other.relation 
                                OR waiting.transactionid = other.transactionid
                            JOIN pg_stat_activity other_stm ON other_stm.pid = other.pid
                            WHERE NOT waiting.granted 
                                AND waiting.pid <> other.pid
                                AND age(clock_timestamp(), waiting_stm.xact_start) > interval %s
                                AND other.pid = %d
                        )""" % (CSR.sys_conf.cancel_blocker_tx_timeout, pid)
                    )
                    if pid_is_locker:
                        db_conn.execute("SELECT pg_cancel_backend(%d)" % pid)
                        CSR.lock.acquire()
                        if pid in CSR.workers_db_pid: CSR.workers_db_pid.remove(pid)
                        CSR.lock.release()
                        CSR.logger.log('%s: stopped pid %d' % (thread_name, pid), "Info")

                CSR.logger.log(
                    '%s: iteration done. Sleep on %d seconds...' %
                    (thread_name, CSR.sys_conf.lock_observer_sleep_interval),
                    "Info"
                )

                sleep_lo()
        except (
            postgresql.exceptions.QueryCanceledError,
            postgresql.exceptions.AdminShutdownError,
            postgresql.exceptions.CrashShutdownError,
            postgresql.exceptions.ServerNotReadyError
        ) as e:
            if CSR.is_terminate:
                return
            do_work = True
            CSR.logger.log(
                'Exception in \'%s\': %s. Reconnecting after %d sec...' %
                    (thread_name, str(e), CSR.sys_conf.conn_exception_sleep_interval),
                "Error"
            )
            time.sleep(CSR.sys_conf.conn_exception_sleep_interval)
        except Exception:
            CSR.logger.log('Exception in %s: \n%s' % (thread_name, exception_helper()), "Error")
        finally:
            db_conn.close()
            CSR.logger.log('Thread %s finished!' % thread_name, "Info")


def worker_idxs_func(thread_name, db_conn_str, db_name, reindex_scenario):
    def dbg_print_queue(elems, name):
        CSR.logger.log(name + " >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> ", "Debug")
        if type(elems) is list:
            for elem in elems:
                CSR.logger.log(str([
                    elem[FldsEnumBasicList.schema_name],
                    elem[FldsEnumBasicList.tbl_name],
                    elem[FldsEnumBasicList.idx_name]
                ]), "Debug")
        if type(elems) is dict:
            for id, elem in elems.items():
                CSR.logger.log(str([
                    id,
                    elem[FldsEnumBasicList.schema_name],
                    elem[FldsEnumBasicList.tbl_name],
                    elem[FldsEnumBasicList.idx_name]
                ]), "Debug")
        CSR.logger.log(name + " <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< ", "Debug")

    def check_access_share_lock(db_conn, thread_name, indrelid):
        access_share_lock = True
        while access_share_lock:
            access_share_lock = get_scalar(
                db_conn,
                """
                SELECT exists(
                    SELECT 1
                    FROM pg_locks
                    WHERE database = (
                        SELECT oid
                        FROM pg_database
                        WHERE datname = current_database()
                    )
                    AND relation = %d
                )""" % indrelid)
            if access_share_lock:
                CSR.logger.log(
                    "%s: access_share_lock found on relation %d! sleep 5 seconds" % (thread_name, indrelid),
                    "Info"
                )
                time.sleep(CSR.sys_conf.a_share_lock_sleep_interval)

    def get_next_idx():
        CSR.lock.acquire()
        for indexrelid, elem in CSR.idxs.items():
            tbl = '%s.%s' % (elem[FldsEnumBasicList.schema_name], elem[FldsEnumBasicList.tbl_name])
            all_active_linked_tbls = list(itertools.chain(*[tbl_i for _, tbl_i in CSR.active_rels.items()]))
            elem_linked_tbls = []
            if elem[FldsEnumBasicList.complex]:
                elem_linked_tbls = list(set([
                        '%s.%s' % (e[FldsEnumComplex.conn_schema_name], e[FldsEnumComplex.conn_tbl_name])
                        for e in CSR.complex_cases[elem[FldsEnumBasicList.indexrelid]]
                        if e[FldsEnumComplex.contype == 'f']
                    ]))
                if len(list(set(all_active_linked_tbls) & set(elem_linked_tbls))):
                    continue

            if elem[FldsEnumBasicList.indrelid] not in CSR.active_idxs and tbl not in all_active_linked_tbls:
                CSR.active_idxs[elem[FldsEnumBasicList.indrelid]] = elem
                if elem[FldsEnumBasicList.complex]:
                    CSR.active_rels[elem[FldsEnumBasicList.indrelid]] = elem_linked_tbls
                CSR.lock.release()
                return list(elem)
        CSR.lock.release()
        return None

    def cancel_task():
        CSR.lock.acquire()
        if elem is not None and elem[FldsEnumBasicList.indrelid] in CSR.active_idxs:
            # cancel processed task
            del CSR.active_idxs[elem[FldsEnumBasicList.indrelid]]
            del CSR.active_rels[elem[FldsEnumBasicList.indrelid]]
        CSR.lock.release()

    def finish_task(task_already_done=False):
        CSR.lock.acquire()
        if CSR.args.track_actions and not task_already_done and 'new_indexrelid' in locals():
            insert_action(
                db_local,
                reindex_scenario,
                elem[FldsEnumBasicList.relnamespace],
                elem[FldsEnumBasicList.indrelid],
                new_indexrelid
            )
        del CSR.active_idxs[elem[FldsEnumBasicList.indrelid]]
        if elem[FldsEnumBasicList.indrelid] in CSR.active_rels:
            del CSR.active_rels[elem[FldsEnumBasicList.indrelid]]
        del CSR.idxs[elem[FldsEnumBasicList.indexrelid]]
        if elem[FldsEnumBasicList.indexrelid] in CSR.atomic_stages:
            del CSR.atomic_stages[elem[FldsEnumBasicList.indexrelid]]
        if elem[FldsEnumBasicList.indexrelid] in CSR.complex_cases:
            del CSR.complex_cases[elem[FldsEnumBasicList.indexrelid]]
        CSR.processed_idx += 1
        CSR.lock.release()

    def _execute(query):
        if CSR.sys_conf.log_sql == 1:
            CSR.logger.log("%s %d: %s" % (thread_name, current_pid, query), "Info")
        if CSR.sys_conf.execute_sql:
            db_local.execute(query)

    def get_grants(db_conn, schema_name, obj_name):
        result = get_resultset(db_conn, """
            SELECT format('GRANT %s ON %s %I.%I TO %I',
                T.privileges,
                T.obj,
                quote_ident('""" + schema_name + """'),
                quote_ident(T.relname),
                T.grantee
            )
            FROM (
                SELECT
                    b.relname,
                    case
                        when b.relkind in ('r', 'v', 'm', 'f') then 'TABLE'
                        when b.relkind in ('S') then 'SEQUENCE'
                    else 'todo'
                    end as obj,
                    COALESCE(gt.rolname, 'PUBLIC') AS grantee,
                    string_agg(b.privilege_type, ', ') AS privileges
                FROM
                    (SELECT
                        (d).grantee AS grantee, (d).grantor AS grantor,
                        (d).is_grantable AS is_grantable,
                        (d).privilege_type, a.relname, a.relkind
                    FROM
                        (SELECT c.relname, c.relkind, aclexplode(c.relacl) as d
                        FROM pg_namespace nsp
                        JOIN pg_class c ON nsp.oid = c.relnamespace
                        WHERE relkind in ('r', 'm', 'v', 'S', 'f')
                            AND nsp.nspname = quote_ident('""" + schema_name + """')
                            AND c.relname = quote_ident('""" + obj_name + """')
                        ) a
                    ) b
                    LEFT JOIN pg_catalog.pg_roles g ON (b.grantor = g.oid)
                    LEFT JOIN pg_catalog.pg_roles gt ON (b.grantee = gt.oid)
                GROUP BY b.relname, b.relkind, g.rolname, gt.rolname
                ORDER BY b.relname, b.relkind, g.rolname
            ) T
        """)
        dlcs = ""
        for v in result:
            dlcs += v[0] + ";\n"
        return dlcs

    def get_dependent_objs(db_conn, schema_name, tbl_name, conn_name):
        result = get_resultset(db_conn, """
            SELECT
                (SELECT 'DROP VIEW ' || quote_ident(n.nspname)
                    || '.' || quote_ident(c.relname)) as drop_ddl,
                (SELECT 'CREATE OR REPLACE VIEW ' || quote_ident(n.nspname)
                    || '.' || quote_ident(c.relname) || ' AS ' || pg_get_viewdef(c.oid, true)) as create_ddl,
                quote_ident(n.nspname), quote_ident(c.relname)
            FROM pg_depend d
            JOIN pg_constraint cn on cn.oid = d.objid and d.classid::regclass = 'pg_constraint'::regclass
            JOIN pg_depend d2 on d2.refobjid = cn.oid and d2.classid::regclass = 'pg_rewrite'::regclass
            JOIN pg_rewrite rw on rw.oid = d2.objid
            JOIN pg_class c on c.oid = rw.ev_class and c.relkind = 'v'
            JOIN pg_namespace n on n.oid = c.relnamespace
            WHERE
                d.refobjid='%s.%s'::regclass
                and cn.conname = '%s'
            GROUP BY n.nspname, c.relname, c.oid, d.refobjid, cn.conname
            """ % (schema_name, tbl_name, conn_name)
        )
        drop_ddls = ""
        create_ddls = ""
        restore_dlcs = ""
        for v in result:
            drop_ddls += v[0] + ";\n"
            create_ddls += v[1] + "\n"
            restore_dlcs += get_grants(db_conn, v[2], v[3])

        return drop_ddls, create_ddls, restore_dlcs

    def count_dependent_objs(db_conn, _elem):
        dependent_objs = []
        if _elem[FldsEnumBasicList.complex]:
            for _complex_item in CSR.complex_cases[elem[FldsEnumBasicList.indexrelid]]:
                if _complex_item[FldsEnumComplex.contype] == 'p':
                    count = get_scalar(db_conn, """
                        SELECT
                            count(1)
                        FROM (
                            SELECT 1
                            FROM pg_depend d
                            JOIN pg_constraint cn on cn.oid = d.objid
                                and d.classid::regclass = 'pg_constraint'::regclass
                            JOIN pg_depend d2 on d2.refobjid = cn.oid
                                and d2.classid::regclass = 'pg_rewrite'::regclass
                            JOIN pg_rewrite rw on rw.oid = d2.objid
                            JOIN pg_class c on c.oid = rw.ev_class and c.relkind = 'v'
                            JOIN pg_namespace n on n.oid = c.relnamespace
                            WHERE
                                d.refobjid='%s.%s'::regclass
                                and cn.conname = '%s'
                            GROUP BY n.nspname, c.relname, c.oid, d.refobjid, cn.conname
                        ) T
                        """ % (
                            _complex_item[FldsEnumComplex.conn_schema_name],
                            _complex_item[FldsEnumComplex.conn_tbl_name],
                            _complex_item[FldsEnumComplex.conn_name]
                        )
                    )
                    dependent_objs.append(int(count if count is not None else 0))
        return 0 if len(dependent_objs) == 0 else max(dependent_objs)

    def is_exists_atomic_stage(indexrelid, stage_type, hash):
        if indexrelid in CSR.atomic_stages:
            for item in CSR.atomic_stages[indexrelid]:
                if item[0] == stage_type and item[1] == hash:
                    CSR.logger.log(
                        "Atomic stage already exists: indexrelid = %d, stage_type = %s, hash = %s, query = %s " % (
                            indexrelid,
                            stage_type,
                            hash,
                            item[2]
                        ), "Info", do_print=True
                    )
                    return True
        return False

    def add_atomic_stage(indexrelid, stage_type, hash, q):
        CSR.atomic_stages.setdefault(indexrelid, []).append((stage_type, hash, q))

    def prepare_session(conn):
        is_super = get_scalar(conn, 'select usesuper from pg_user where usename = CURRENT_USER')
        try:
            if CSR.sys_conf.log_level == "Debug":
                conn.execute("SET log_min_error_statement = 'INFO'")
                conn.execute("SET log_min_duration_statement = 0")
                conn.execute("SET log_lock_waits = on")
                conn.execute("SET log_statement = 'all'")
                if is_super:
                    conn.execute("SET deadlock_timeout = '1h'")
            else:
                if is_super:
                    conn.execute("SET deadlock_timeout = '%s'" % getattr(CSR.sys_conf, 'deadlock_timeout'))
                conn.execute("SET statement_timeout = '%s'" % getattr(CSR.sys_conf, 'statement_timeout'))
                conn.execute("SET work_mem = '%s'" % getattr(CSR.sys_conf, 'work_mem'))
                conn.execute("SET maintenance_work_mem = '%s'" % getattr(CSR.sys_conf, 'maintenance_work_mem'))
        except (
            postgresql.exceptions.InsufficientPrivilegeError
        ) as e:
            CSR.logger.log(
                'Exception in prepare_session: %s. Skip configuring session variables...' % str(e),
                "Error"
            )

    CSR.logger.log("Started '%s' thread for '%s' database" % (thread_name, db_name), "Info")

    elem = None
    do_work = True
    db_local = None
    current_pid = None

    while do_work:
        do_work = False
        try:
            # check conn to DB
            if db_local is not None:    # connection db_local has been already established
                CSR.logger.log("Thread '%s': try 'SELECT 1'" % thread_name, "Info")
                try:
                    db_local.execute("SELECT 1")    # live?
                except:
                    CSR.logger.log("Thread '%s': Connection to DB is broken" % thread_name, "Error", do_print=True)
                    db_local = None     # needs reconnect

            # ======================================================
            # connecting to DB, session variables initialization
            if db_local is None:
                if current_pid is not None and current_pid in CSR.workers_db_pid:
                    CSR.workers_db_pid.remove(current_pid)

                CSR.logger.log("Thread '%s': connecting to '%s' database..." % (thread_name, db_name), "Info")
                db_local = postgresql.open(db_conn_str)
                db_local.execute(
                    "SET application_name = '%s'" % (
                            CSR.sys_conf.application_name + "_" + os.path.splitext(reindex_scenario)[0]
                    )
                )

                prepare_session(db_local)

                current_pid = get_scalar(db_local, "SELECT pg_backend_pid()")
                CSR.db_conns[current_pid] = db_local
                CSR.workers_db_pid.append(current_pid)
                CSR.logger.log(
                    "Thread '%s': connected to '%s' database with pid %d" %
                    (thread_name, db_name, current_pid),
                    "Info"
                )
            # ======================================================

            while len(CSR.idxs) > 0:
                time.sleep(0.01)

                if len(CSR.idxs) == 0:
                    break

                elem = get_next_idx()
                if elem is None:
                    CSR.logger.log(
                        "%s %d: all tables in process, waiting 5 sec..." % (thread_name, current_pid),
                        "Info",
                        do_print=True
                    )
                    time.sleep(CSR.sys_conf.a_share_lock_sleep_interval)
                    continue

                if [
                    elem[FldsEnumBasicList.schema_name],
                    elem[FldsEnumBasicList.tbl_name],
                    elem[FldsEnumBasicList.idx_name]
                ] in CSR.ban_idxs:
                    CSR.logger.log(
                        "%s %d: element %s found in ban list" % (thread_name, current_pid, str(elem)),
                        "Info",
                        do_print=True
                    )
                    finish_task()
                    continue

                if CSR.args.track_actions and is_action_exists(
                        db_local,
                        reindex_scenario,
                        elem[FldsEnumBasicList.relnamespace],
                        elem[FldsEnumBasicList.indrelid],
                        elem[FldsEnumBasicList.indexrelid]
                ):
                    CSR.logger.log(
                        "%s %d: element %s already exists" % (thread_name, current_pid, str(elem)),
                        "Info",
                        do_print=True
                    )
                    finish_task(task_already_done=True)
                    continue

                # ==============================================================================
                if elem[FldsEnumBasicList.complex] and CSR.args.skip_fk:
                    finish_task()
                    CSR.logger.log(
                        "%s: skipped complex case %s, processed %d from %d =====> %s" % (
                            thread_name,
                            str(elem),
                            CSR.processed_idx,
                            CSR.total_idxs,
                            str(round(float(CSR.processed_idx * 100 / CSR.total_idxs), 1)) + "%"
                        ), "Info", do_print=True
                    )
                    continue
                # ==============================================================================

                elem_orig = list(elem)

                if CSR.sys_conf.log_level == "Debug":
                    dbg_print_queue(CSR.idxs, "idxs in %s" % thread_name)
                    dbg_print_queue(CSR.active_idxs, "active_idxs in %s" % thread_name)

                count_dependent_objs_res = count_dependent_objs(db_local, elem_orig)
                if count_dependent_objs_res > CSR.sys_conf.max_dependent_objects:
                    finish_task()
                    CSR.logger.log(
                        """%s: skipped complex case %s,
                        reason dependent objects is %d (max_dependent_objects = %d),
                        processed %d from %d =====> %s""" % (
                            thread_name,
                            str(elem),
                            count_dependent_objs_res,
                            CSR.sys_conf.max_dependent_objects,
                            CSR.processed_idx,
                            CSR.total_idxs,
                            str(round(float(CSR.processed_idx * 100 / CSR.total_idxs), 1)) + "%"
                        ), "Info", do_print=True
                    )
                    continue

                query = "SET search_path = '%s'" % elem[FldsEnumBasicList.schema_name]
                _execute(query)

                if len(elem[FldsEnumBasicList.idx_name]) > 63 - len(CSR.sys_conf.postfix_str):
                    new_elem_name = str(
                        elem[FldsEnumBasicList.idx_name]
                    )[0:63 - len(CSR.sys_conf.postfix_str)] + CSR.sys_conf.postfix_str
                else:
                    new_elem_name = str(elem[FldsEnumBasicList.idx_name]) + CSR.sys_conf.postfix_str

                # ==========================================================
                # try remove undone idx
                q_drop_index = """
                    DO $$
                    BEGIN
                        IF EXISTS (
                            SELECT 1
                            FROM   pg_class c
                            JOIN   pg_namespace n ON n.oid = c.relnamespace
                            WHERE  c.relname = '%s' AND c.relkind = 'i' AND n.nspname = '%s'
                         ) THEN
                            DROP INDEX if exists "%s"."%s";
                        END IF;
                    END$$;""" % (
                            new_elem_name,
                            elem[FldsEnumBasicList.schema_name],
                            elem[FldsEnumBasicList.schema_name],
                            new_elem_name
                        )
                if not is_exists_atomic_stage(
                    elem[FldsEnumBasicList.indexrelid],
                    AtomicStageType.INDEX_CREATE.value,
                    hashlib.md5(q_drop_index.encode()).hexdigest()
                ):
                    check_access_share_lock(db_local, thread_name, elem[FldsEnumBasicList.indrelid])
                    _execute(q_drop_index)
                    add_atomic_stage(
                        elem[FldsEnumBasicList.indexrelid],
                        AtomicStageType.INDEX_CREATE.value,
                        hashlib.md5(q_drop_index.encode()).hexdigest(),
                        q_drop_index
                    )
                # ==========================================================
                # CREATE index
                q_create_index = elem_orig[FldsEnumBasicList.idx_def]
                q_create_index = q_create_index.replace(elem_orig[FldsEnumBasicList.idx_name], new_elem_name)
                q_create_index = q_create_index.replace("INDEX ", "INDEX concurrently ")
                if not is_exists_atomic_stage(
                    elem[FldsEnumBasicList.indexrelid],
                    AtomicStageType.INDEX_CREATE.value,
                    hashlib.md5(q_create_index.encode()).hexdigest()
                ):
                    _execute(q_create_index)
                    add_atomic_stage(
                        elem[FldsEnumBasicList.indexrelid],
                        AtomicStageType.INDEX_CREATE.value,
                        hashlib.md5(q_create_index.encode()).hexdigest(),
                        q_create_index
                    )

                if elem[FldsEnumBasicList.complex]:
                    # ================================================================================
                    # complex case: FKs
                    # ================================================================================
                    complex_items = CSR.complex_cases[elem[FldsEnumBasicList.indexrelid]]

                    # drop stage
                    for complex_item in complex_items:
                        if complex_item[FldsEnumComplex.contype] == 'f':
                            check_access_share_lock(db_local, thread_name, elem[FldsEnumBasicList.indrelid])
                            q_drop_constr = """
                                ALTER TABLE "%s"."%s" DROP CONSTRAINT IF EXISTS "%s";
                            """ % (
                                complex_item[FldsEnumComplex.conn_schema_name],
                                complex_item[FldsEnumComplex.conn_tbl_name],
                                complex_item[FldsEnumComplex.conn_name]
                            )
                            if not is_exists_atomic_stage(
                                elem[FldsEnumBasicList.indexrelid],
                                AtomicStageType.FK_DROP.value,
                                hashlib.md5(q_drop_constr.encode()).hexdigest()
                            ):
                                _execute(q_drop_constr)
                                add_atomic_stage(
                                    elem[FldsEnumBasicList.indexrelid],
                                    AtomicStageType.FK_DROP.value,
                                    hashlib.md5(q_drop_constr.encode()).hexdigest(),
                                    q_drop_constr
                                )

                    # replace stage
                    for complex_item in complex_items:
                        if complex_item[FldsEnumComplex.contype] in ('p', 'u'):
                            if complex_item[FldsEnumComplex.contype] == 'p':
                                depended_objs = get_dependent_objs(
                                    db_local,
                                    complex_item[FldsEnumComplex.conn_schema_name],
                                    complex_item[FldsEnumComplex.conn_tbl_name],
                                    complex_item[FldsEnumComplex.conn_name]
                                )
                            else:
                                depended_objs = "", "", ""

                            check_access_share_lock(db_local, thread_name, elem[FldsEnumBasicList.indrelid])

                            # drop ddls
                            q_replace_constr = depended_objs[0] + "\n" + """
                                    ALTER TABLE "%s"."%s"
                                      DROP CONSTRAINT IF EXISTS "%s",
                                      ADD CONSTRAINT "%s" %s USING INDEX "%s";
                                """ % (
                                    complex_item[FldsEnumComplex.conn_schema_name],
                                    complex_item[FldsEnumComplex.conn_tbl_name],
                                    complex_item[FldsEnumComplex.conn_name],
                                    complex_item[FldsEnumComplex.conn_name],
                                    'PRIMARY KEY' if complex_item[FldsEnumComplex.contype] == 'p' else 'UNIQUE',
                                    new_elem_name
                                ) + "\n" + depended_objs[1] + "\n" + depended_objs[2]

                            if not is_exists_atomic_stage(
                                elem[FldsEnumBasicList.indexrelid],
                                AtomicStageType.FK_USING_INDEX_ADD.value,
                                hashlib.md5(q_replace_constr.encode()).hexdigest()
                            ):
                                _execute(q_replace_constr)
                                add_atomic_stage(
                                    elem[FldsEnumBasicList.indexrelid],
                                    AtomicStageType.FK_USING_INDEX_ADD.value,
                                    hashlib.md5(q_replace_constr.encode()).hexdigest(),
                                    q_replace_constr
                                )

                    # restore stage
                    for complex_item in complex_items:
                        if complex_item[FldsEnumComplex.contype] == 'f':
                            q_add_fk_not_valid = """
                                ALTER TABLE "%s"."%s"
                                    ADD CONSTRAINT "%s" %s%s;
                            """ % (
                                complex_item[FldsEnumComplex.conn_schema_name],
                                complex_item[FldsEnumComplex.conn_tbl_name],
                                complex_item[FldsEnumComplex.conn_name],
                                complex_item[FldsEnumComplex.condef],
                                ' NOT VALID' if complex_item[FldsEnumComplex.condef].find('NOT VALID') == -1 else ''
                            )
                            if not is_exists_atomic_stage(
                                elem[FldsEnumBasicList.indexrelid],
                                AtomicStageType.FK_NOT_VALID_ADD.value,
                                hashlib.md5(q_add_fk_not_valid.encode()).hexdigest()
                            ):
                                _execute(q_add_fk_not_valid)
                                add_atomic_stage(
                                    elem[FldsEnumBasicList.indexrelid],
                                    AtomicStageType.FK_NOT_VALID_ADD.value,
                                    hashlib.md5(q_add_fk_not_valid.encode()).hexdigest(),
                                    q_add_fk_not_valid
                                )

                            q_validate_fk = """
                                ALTER TABLE "%s"."%s"
                                    VALIDATE CONSTRAINT "%s";
                            """ % (
                                complex_item[FldsEnumComplex.conn_schema_name],
                                complex_item[FldsEnumComplex.conn_tbl_name],
                                complex_item[FldsEnumComplex.conn_name]
                            )
                            if not is_exists_atomic_stage(
                                elem[FldsEnumBasicList.indexrelid],
                                AtomicStageType.FK_VALIDATE.value,
                                hashlib.md5(q_validate_fk.encode()).hexdigest()
                            ):
                                _execute(q_validate_fk)
                                add_atomic_stage(
                                    elem[FldsEnumBasicList.indexrelid],
                                    AtomicStageType.FK_VALIDATE.value,
                                    hashlib.md5(q_validate_fk.encode()).hexdigest(),
                                    q_validate_fk
                                )
                    # ================================================================================

                if not elem[FldsEnumBasicList.complex]:
                    # DROP index
                    check_access_share_lock(db_local, thread_name, elem[FldsEnumBasicList.indrelid])
                    q_replace_index = """
                        DROP INDEX "%s"."%s";
                        ALTER INDEX "%s"."%s" RENAME TO "%s";
                    """ % (
                        elem_orig[FldsEnumBasicList.schema_name],
                        elem_orig[FldsEnumBasicList.idx_name],
                        elem_orig[FldsEnumBasicList.schema_name],
                        new_elem_name,
                        elem_orig[FldsEnumBasicList.idx_name]
                    )
                    if not is_exists_atomic_stage(
                            elem[FldsEnumBasicList.indexrelid],
                            AtomicStageType.INDEX_CREATE.value,
                            hashlib.md5(q_replace_index.encode()).hexdigest()
                    ):
                        _execute(q_replace_index)
                        add_atomic_stage(
                            elem[FldsEnumBasicList.indexrelid],
                            AtomicStageType.INDEX_CREATE.value,
                            hashlib.md5(q_replace_index.encode()).hexdigest(),
                            q_replace_index
                        )

                new_indexrelid = get_scalar(db_local, """
                    SELECT i.indexrelid
                    FROM pg_class c
                    INNER JOIN pg_index i on c.oid = i.indexrelid
                    INNER JOIN pg_namespace n on n.oid = c.relnamespace
                    WHERE n.nspname = '%s'
                        AND c.relname = '%s'
                        AND c.relkind = 'i'
                        AND i.indrelid = %d
                """ % (
                    elem[FldsEnumBasicList.schema_name],
                    elem[FldsEnumBasicList.idx_name],
                    elem[FldsEnumBasicList.indrelid]
                ))
                # ======================================================================
                # release active table and remove processed "elem" from "CSR.idxs"
                finish_task()
                # ======================================================================
                if CSR.sys_conf.log_level == "Debug":
                    dbg_print_queue(CSR.active_idxs, "active_idxs after del in %s" % thread_name)
                CSR.logger.log(
                    "%s: finish operation, processed %d from %d =====> %s" % (
                        thread_name,
                        CSR.processed_idx,
                        CSR.total_idxs,
                        str(round(float(CSR.processed_idx * 100 / CSR.total_idxs), 1)) + "%"
                    ), "Info", do_print=True
                )
        except (
            postgresql.exceptions.UndefinedObjectError,
            postgresql.exceptions.UndefinedTableError,
            postgresql.exceptions.DuplicateObjectError,
            postgresql.exceptions.SyntaxError
        ) as e:
            do_work = True
            CSR.logger.log(
                'Exception in \'%s\', PID = %s:\n%s' % (
                    thread_name, str(current_pid), str(e)
                ),
                "Error",
                do_print=True
            )
            # ======================================================================
            # release active table and remove processed "elem" from "CSR.idxs"
            finish_task()
            # ======================================================================
        except postgresql.exceptions.DeadlockError:
            do_work = True
            CSR.logger.log(
                'Exception in \'%s\', PID = %s: Deadlock detected' % (
                    thread_name, str(current_pid)
                ),
                "Error",
                do_print=True
            )
            cancel_task()
            time.sleep(CSR.sys_conf.conn_exception_sleep_interval)
        except AttributeError:
            do_work = True
            CSR.logger.log(
                'Exception in \'%s\', PID = %s: AttributeError %s' % (
                    thread_name, str(current_pid), exception_helper()
                ),
                "Error",
                do_print=True
            )
            cancel_task()
            time.sleep(CSR.sys_conf.conn_exception_sleep_interval)
        except postgresql.exceptions.UniqueError:
            do_work = True
            CSR.logger.log(
                'Exception in \'%s\', PID = %s: UniqueError detected' % (
                    thread_name, str(current_pid)
                ),
                "Error",
                do_print=True
            )
            time.sleep(CSR.sys_conf.conn_exception_sleep_interval)
            # ===============================================================================
            try:
                for rule_name, rule in CSR.rules.items():
                    if rule["except_type"] == "UniqueError" and rule["tbl_name"] == elem[FldsEnumBasicList.tbl_name] and \
                            elem[FldsEnumBasicList.idxdef].find(rule["idxdef_like"]) > -1 and \
                            elem[FldsEnumBasicList.indisunique]:
                        CSR.logger.log(
                            "Found match with rule %s, elem %s" % (rule_name, str(elem)),
                            "Info",
                            do_print=True
                        )
                        for cmd in rule["cmds"]:
                            cmd = cmd.replace("BACKUP_DIR", CSR.backup_dir)
                            cmd = cmd.replace("SCHEMA_NAME", elem[FldsEnumBasicList.schema_name])
                            if CSR.sys_conf.log_sql == 1:
                                CSR.logger.log("%s %d: %s" % (thread_name, current_pid, cmd), "Info", do_print=True)
                            if CSR.sys_conf.execute_sql:
                                if cmd.upper().find("COPY") > -1 and cmd.upper().find("GZIP") > -1:
                                    output_files = re.findall(r"gzip\s\>\s(.*?)\.gz", cmd)
                                    existed_files = [f for f in output_files if os.path.isfile(f + '.gz')]

                                    # NOTE: this logic works only if custom_soft_reindex runned on same host as DB.
                                    # Re-backup with overwriting of files is possible in case when the unique index
                                    # was not built from the first attempt after correction of the data, that is,
                                    # during the modification time, not-unique values were recorded.
                                    # This is possible only if the unique index is corrupted.

                                    if len(output_files) == 0 or len(output_files) != len(existed_files):
                                        CSR.logger.log(
                                            "%s: backup files %s" % (thread_name, str(output_files)),
                                            "Info",
                                            do_print=True
                                        )
                                        with db_local.xact(isolation='REPEATABLE READ'):
                                            db_local.execute(cmd)
                                    else:
                                        CSR.logger.log(
                                            "%s: files %s already exists" % (thread_name, str(existed_files)),
                                            "Info",
                                            do_print=True
                                        )
                                else:
                                    db_local.execute(cmd)
            except (
                    postgresql.exceptions.QueryCanceledError,
                    postgresql.exceptions.AdminShutdownError,
                    postgresql.exceptions.CrashShutdownError,
                    postgresql.exceptions.ServerNotReadyError
            ) as e:
                if CSR.is_terminate:
                    return
                do_work = True
                CSR.logger.log(
                    'Exception in \'%s\', PID = %s: %s on UniqueError processing. Reconnecting after 5 sec...' % (
                        thread_name, str(current_pid), str(e)
                    ),
                    "Error",
                    do_print=True
                )
                cancel_task()
                db_local = None
                time.sleep(CSR.sys_conf.conn_exception_sleep_interval)
            except:
                do_work = True
                CSR.logger.log(
                    'Exception in \'%s\', PID = %s on UniqueError processing: \n%s' % (
                        thread_name, str(current_pid), exception_helper()
                    ),
                    "Error",
                    do_print=True
                )
                cancel_task()
                time.sleep(CSR.sys_conf.conn_exception_sleep_interval)
            # ===============================================================================
            cancel_task()
        except (
            postgresql.exceptions.QueryCanceledError,
            postgresql.exceptions.AdminShutdownError,
            postgresql.exceptions.CrashShutdownError,
            postgresql.exceptions.ServerNotReadyError
        ) as e:
            if CSR.is_terminate:
                return
            do_work = True
            CSR.logger.log(
                'Exception in \'%s\', PID = %s: %s. Reconnecting after %d sec...' % (
                    thread_name, str(current_pid), str(e), CSR.sys_conf.conn_exception_sleep_interval
                ),
                "Error",
                do_print=True
            )
            cancel_task()
            db_local = None
            time.sleep(CSR.sys_conf.conn_exception_sleep_interval)
        except:
            if CSR.is_terminate:
                return
            do_work = False
            CSR.logger.log(
                'Exception in \'%s\', PID = %s: %s' % (
                    thread_name, str(current_pid), exception_helper()
                ),
                "Error",
                do_print=True
            )
            cancel_task()

    CSR.lock.acquire()  # give peace of mind
    if current_pid is not None and current_pid in CSR.workers_db_pid:
        CSR.workers_db_pid.remove(current_pid)
    CSR.lock.release()
    CSR.logger.log(
        "Finished '%s', PID = %s for '%s' database" % (
            thread_name, str(current_pid), db_name
        ), "Info"
    )


def get_idxs(str_conn, reindex_scenario):
    sql_query_file = open(
        os.path.join(
            CSR.sys_conf.current_dir,
            'reindex_scenarios',
            reindex_scenario
        ),
        'r'
    )
    sql_query = sql_query_file.read()
    sql_query_file.close()
    db_local = postgresql.open(str_conn)
    db_local.execute(
            "SET application_name = '%s'" %
            (CSR.sys_conf.application_name + "_" + os.path.splitext(reindex_scenario)[0])
    )
    sql_query = sql_query.replace(
        "--EXT_COND",
        "AND ci.relname not ilike '%s'" % ('%' + CSR.sys_conf.postfix_str)
    )
    if CSR.sys_conf.log_sql == 1:
        CSR.logger.log("get_idxs: %s" % sql_query, "Info")
    result = get_resultset(db_local, sql_query)
    db_local.close()
    return result


def init_action_tracker_tbls(str_conn):
    db_conn = postgresql.open(str_conn)
    db_conn.execute("""
    DO $$
        begin
        IF not EXISTS (
             SELECT 1
             FROM   pg_class c
             JOIN   pg_namespace n ON n.oid = c.relnamespace
             WHERE  c.relname = 'csr_actions'
             AND    n.nspname = 'public'
        ) OR not EXISTS (
             SELECT 1
             FROM   pg_class c
             JOIN   pg_namespace n ON n.oid = c.relnamespace
             WHERE  c.relname = 'csr_scenarios'
             AND    n.nspname = 'public'
        )
        THEN

            DROP TABLE IF EXISTS public.csr_scenarios CASCADE;
            DROP TABLE IF EXISTS public.csr_actions CASCADE;

            CREATE TABLE public.csr_scenarios
            (
                id serial,
                name character varying(128),
                CONSTRAINT csr_scenarios_pkey PRIMARY KEY (id)
            )
            WITH (
              OIDS=FALSE
            );

            CREATE TABLE public.csr_actions
            (
                dt timestamp with time zone default now(),
                reindex_scenario_id integer,
                relnamespace oid,
                indrelid oid,
                indexrelid oid,
                CONSTRAINT csr_actions_uniq UNIQUE (indexrelid, indrelid, relnamespace, reindex_scenario_id),
                CONSTRAINT fk_csr_scenarios_id FOREIGN KEY (reindex_scenario_id)
                    REFERENCES public.csr_scenarios (id)
            )
            WITH (
              OIDS=FALSE
            );

            CREATE INDEX csr_actions_dt_idx
                ON public.csr_actions USING btree (dt);
        END IF;
        end$$;
    """)
    db_conn.close()


def is_action_exists(db_conn, reindex_scenario, relnamespace, indrelid, indexrelid):
    _q = """
            SELECT EXISTS(
                SELECT 1
                FROM public.csr_actions a
                JOIN public.csr_scenarios s ON a.reindex_scenario_id = s.id
                WHERE s.name = '%s'
                    AND a.relnamespace = %d
                    AND a.indrelid = %d
                    AND a.indexrelid = %d
                    %s
            )
        """ % (
            reindex_scenario,
            relnamespace,
            indrelid,
            indexrelid, (
                '' if CSR.args.action_expiration_hours is None or CSR.args.action_expiration_hours == 0 else
                "AND a.dt > now() - interval '%d hours'" % CSR.args.action_expiration_hours
            )
        )
    CSR.logger.log(_q, "Debug")
    return get_scalar(db_conn, _q)


def insert_action(db_conn, reindex_scenario, relnamespace, indrelid, indexrelid):
    _q = """
        DO $$
        declare
            scenario_name text = '%s';
            scenario_id integer = NULL;
        begin
            SELECT id into scenario_id
            FROM   public.csr_scenarios
            WHERE  name = scenario_name;

             IF scenario_id IS NULL THEN
                 INSERT INTO public.csr_scenarios(name)
                    VALUES (scenario_name) RETURNING id INTO scenario_id;
             END IF;

             %s

             INSERT INTO public.csr_actions(reindex_scenario_id, relnamespace, indrelid, indexrelid)
                VALUES (scenario_id, %d, %d, %d);
        end$$;
    """ % (
            reindex_scenario,
            '' if CSR.args.action_expiration_hours is None or CSR.args.action_expiration_hours == 0 else
            """
                DELETE FROM public.csr_actions
                WHERE
                    reindex_scenario_id = scenario_id
                    AND relnamespace = %d
                    AND indrelid = %d
                    AND indexrelid = %d;
            """ % (relnamespace, indrelid, indexrelid),
            relnamespace,
            indrelid,
            indexrelid
    )
    CSR.logger.log(_q, "Debug")
    db_conn.execute(_q)


def get_idxs_integrity(str_conn, reindex_scenario):
    sql_query = """
        select
            n.nspname,
            cr.relname,
            ci.relname,
            i.indisprimary,
            i.indisunique,
            (SELECT pg_get_indexdef (i.indexrelid)) as def,
            cn.contype,
            (select pg_get_constraintdef(cn.oid)) as constraintdef,
            nc.nspname,
            cc.relname,
            cn.conname
        from pg_index i
        JOIN pg_class AS ci ON i.indexrelid = ci.oid and ci.relkind = 'i'
        JOIN pg_class AS cr ON i.indrelid = cr.oid and cr.relkind in ('r', 'm')
        JOIN pg_namespace n ON n.oid = ci.relnamespace
        LEFT JOIN pg_constraint cn ON cn.connamespace = n.oid and cn.conindid = i.indexrelid
        LEFT JOIN pg_class AS cc ON cn.conrelid = cc.oid
        LEFT JOIN pg_namespace AS nc ON nc.oid = cc.relnamespace
        WHERE n.nspname not in ('pg_catalog', 'pg_toast') AND ci.relpages > 1 and ci.relname not like '%{}'
        ORDER BY n.nspname, cr.relname, ci.relname;
        """.format(CSR.sys_conf.postfix_str)

    db_local = postgresql.open(str_conn)
    db_local.execute(
            "SET application_name = '%s'" %
            (CSR.sys_conf.application_name + "_" + os.path.splitext(reindex_scenario)[0])
    )
    if CSR.sys_conf.log_sql == 1:
        CSR.logger.log("get_idxs_integrity: %s" % sql_query, "Info")
    result = get_resultset(db_local, sql_query)
    db_local.close()

    return [
        (
            elem[FldsEnumIntegrity.schema_name],
            elem[FldsEnumIntegrity.tbl_name],
            elem[FldsEnumIntegrity.idx_name],
            elem[FldsEnumIntegrity.indisprim],
            elem[FldsEnumIntegrity.indisunique],
            elem[FldsEnumIntegrity.idxdef],
            elem[FldsEnumIntegrity.contype],
            elem[FldsEnumIntegrity.condef],
            elem[FldsEnumIntegrity.conn_schema_name],
            elem[FldsEnumIntegrity.conn_tbl_name],
            elem[FldsEnumIntegrity.conn_name]

        ) for elem in result
    ]


def open_complex_case(str_conn, reindex_scenario, nspname, tbl_name, idx_name):
    sql_query = """
        select
            n.nspname,
            cr.relname,
            ci.relname,
            i.indisprimary,
            i.indisunique,
            cn.contype::text,
            (select pg_get_constraintdef(cn.oid)) as constraintdef,
            nc.nspname,
            cc.relname,
            cn.conname
        from pg_index i
        JOIN pg_class AS ci ON i.indexrelid = ci.oid and ci.relkind = 'i'
        JOIN pg_class AS cr ON i.indrelid = cr.oid and cr.relkind in ('r', 'm')
        JOIN pg_namespace n ON n.oid = ci.relnamespace
        JOIN pg_constraint cn ON /*cn.connamespace = n.oid and*/ cn.conindid = i.indexrelid
        JOIN pg_class AS cc ON cn.conrelid = cc.oid
        JOIN pg_namespace AS nc ON nc.oid = cc.relnamespace
        WHERE n.nspname = '%s' and cr.relname = '%s' and ci.relname = '%s'
        ORDER BY cn.contype, cn.conname;
        """ % (nspname, tbl_name, idx_name)

    db_local = postgresql.open(str_conn)
    db_local.execute(
            "SET application_name = '%s'" %
            (CSR.sys_conf.application_name + "_" + os.path.splitext(reindex_scenario)[0])
    )
    if CSR.sys_conf.log_sql == 1:
        CSR.logger.log("open_complex_case: %s" % sql_query, "Info")
    result = get_resultset(db_local, sql_query)
    db_local.close()

    return [
        (
            elem[FldsEnumComplex.schema_name],
            elem[FldsEnumComplex.tbl_name],
            elem[FldsEnumComplex.idx_name],
            elem[FldsEnumComplex.indisprim],
            elem[FldsEnumComplex.indisunique],
            elem[FldsEnumComplex.contype],
            elem[FldsEnumComplex.condef],
            elem[FldsEnumComplex.conn_schema_name],
            elem[FldsEnumComplex.conn_tbl_name],
            elem[FldsEnumComplex.conn_name]
        ) for elem in result
    ]


def run_test(str_conn, test_name):
    sql_query_file = open(
        os.path.join(
            CSR.sys_conf.current_dir,
            'test',
            test_name
        ), 'r'
    )
    sql_query = sql_query_file.read()
    sql_query_file.close()
    db_local = postgresql.open(str_conn)
    db_local.execute(
            "SET application_name = '%s'" %
            (CSR.sys_conf.application_name + "_" + os.path.splitext(test_name)[0])
    )
    if CSR.sys_conf.log_sql == 1:
        CSR.logger.log("run_test: %s" % sql_query, "Info")
    db_local.execute(sql_query)
    db_local.close()


if __name__ == "__main__":
    def run_on_db(db_name, str_conn):
        global CSR
        CSR.logger.log(
            '=====> CSR %s with scenario \'%s\' started for \'%s\' database!' % (
                VERSION, CSR.args.reindex_scenario, db_name
            ),
            "Info"
        )
        try:
            if CSR.args.track_actions:
                init_action_tracker_tbls(str_conn)

            sc_name = os.path.splitext(CSR.args.reindex_scenario)[0]

            params_str = "Parameters: \n"
            for k in CSR.args.__dict__:
                params_str += '   %s = %s\n' % (k, CSR.args.__dict__[k])

            CSR.logger.log(params_str, "Info", do_print=True)

            CSR.logger.log('Started getting indexes in \'%s\' database...' % db_name, "Info")

            if CSR.args.test is not None and len(CSR.args.test) != 0:
                run_test(str_conn, CSR.args.test)
                time.sleep(5)

            # =======================================================================
            idxs_before = get_idxs(str_conn, CSR.args.reindex_scenario)
            CSR.idxs = {elem[FldsEnumBasicList.indexrelid]: elem for elem in idxs_before}
            for _, elem in CSR.idxs.items():
                if elem[FldsEnumBasicList.complex]:
                    CSR.complex_cases[elem[FldsEnumBasicList.indexrelid]] = \
                        open_complex_case(
                            str_conn,
                            CSR.args.reindex_scenario,
                            elem[FldsEnumBasicList.schema_name],
                            elem[FldsEnumBasicList.tbl_name],
                            elem[FldsEnumBasicList.idx_name]
                        )
            # =======================================================================

            rules_file_name = os.path.join(
                CSR.sys_conf.current_dir,
                'reindex_scenarios',
                'rules',
                sc_name + ".py"
            )
            if os.path.isfile(rules_file_name):
                try:
                    rules_file = open(rules_file_name, 'r')
                    CSR.rules = eval(rules_file.read())
                    rules_file.close()
                except:
                    CSR.logger.log(
                        "Can't load file %s %s" % (rules_file_name, exception_helper()),
                        "Error",
                        do_print=True
                    )

            CSR.logger.log(
                'Found %d idxs in database \'%s\' for current scenario' % (len(CSR.idxs), db_name),
                "Info",
                do_print=True
            )

            idxs_before_integr = get_idxs_integrity(str_conn, sc_name)
            CSR.logger.log(
                'Total related objects in database \'%s\' is %d' % (db_name, len(idxs_before)),
                "Info",
                do_print=True
            )

            CSR.total_idxs = len(CSR.idxs)
            if len(CSR.idxs) > 0:
                worker_threads = []
                CSR.logger.log('Start threads initialization for db = %s' % db_name, "Info")
                for t_num in range(1, CSR.sys_conf.threads_num + 1):
                    worker_threads.append(
                        Thread(
                            target=worker_idxs_func,
                            args=["manager_idx_%s" % str(t_num), str_conn, db_name, sc_name]
                        )
                    )

                worker_threads.append(Thread(target=lock_observer, args=["lock_observer", str_conn, sc_name]))
                for thread in worker_threads: thread.start()
                alive_count = 1
                live_iteration = 0
                CSR.logger.log('Threads successfully initialized for db = %s' % db_name, "Info")
                while alive_count > 0:
                    with SignalHandler() as handler:
                        alive_count = len([thread for thread in worker_threads if thread.is_alive()])
                        if alive_count == 0: break
                        time.sleep(0.5)
                        if live_iteration % (20 * 3) == 0: CSR.logger.log('Live %s threads' % alive_count, "Info")
                        live_iteration += 1
                        if handler.interrupted:
                            CSR.is_terminate = True
                            CSR.logger.log('Received termination signal!', "Info", do_print=True)
                            for _, conn in CSR.db_conns.items():
                                conn.interrupt()
                            if CSR.file_lock is not None:
                                CSR.logger.log('Stopping with AppFileLock...', "Info", do_print=True)
                                handler.unlock()
                            else:
                                CSR.logger.log('Stopping without AppFileLock...', "Info", do_print=True)
                                PSCLogger.instance().stop()
                                sys.exit(0)
            else:
                CSR.logger.log(
                    'Indexes not found in db = %s for scenario %s' % (db_name, sc_name),
                    "Info",
                    do_print=True
                )

            # =======================================================================
            idxs_after = get_idxs(str_conn, CSR.args.reindex_scenario)
            ban_idxs = list(set(idxs_after) - set(idxs_before))
            for idx in ban_idxs:
                if [
                    idx[FldsEnumBasicList.schema_name],
                    idx[FldsEnumBasicList.tbl_name],
                    idx[FldsEnumBasicList.idx_name]
                ] not in CSR.ban_idxs:
                    CSR.ban_idxs.append([
                        idx[FldsEnumBasicList.schema_name],
                        idx[FldsEnumBasicList.tbl_name],
                        idx[FldsEnumBasicList.idx_name]
                    ])
            # =======================================================================

            CSR.logger.log('Finished database \'%s\'' % db_name, "Info", do_print=True)
            CSR.logger.log('Start checking indexes in \'%s\' database...' % db_name, "Info", do_print=True)

            if CSR.args.test_validate is not None and len(CSR.args.test_validate) != 0:
                run_test(str_conn, CSR.args.test_validate)

            idxs_after_integr = get_idxs_integrity(str_conn, sc_name)
            diff_new = list(set(idxs_after_integr) - set(idxs_before_integr))            # only new
            diff_lost = list(set(idxs_before_integr) - set(idxs_after_integr))           # only old

            if len(diff_new) > 0:
                CSR.logger.log(
                    'Integrity violation detected: %d new object(s)' % len(diff_new),
                    "Warning",
                    do_print=True
                )
                for v in diff_new:
                    CSR.logger.log(str(v), "Warning", do_print=True)

            if len(diff_lost) > 0:
                CSR.logger.log(
                    'Integrity violation detected: %d lost object(s)' % len(diff_lost),
                    "Error",
                    do_print=True
                )
                for v in diff_lost:
                    CSR.logger.log(str(v), "Error", do_print=True)

            if len(diff_new) == 0 and len(diff_lost) == 0:
                CSR.logger.log('Integrity checking successfully complete! ', "Info", do_print=True)
        except:
            CSR.logger.log(
                'Exception during processing \'%s\' database: \n%s' % (db_name, exception_helper()),
                "Error",
                do_print=True
            )
        finally:
            CSR.reset()

        CSR.logger.log(
            '<===== CSR %s with scenario \'%s\' finished for \'%s\' database!' % (
                VERSION, CSR.args.reindex_scenario, db_name
            ),
            "Info"
        )

    CSR = CSRGlobal()
    iter_num = 1

    if CSR.args.db_name == 'ALL':
        for db_name, str_conn in CSR.sys_conf.dbs_dict.items():
            run_on_db(db_name, str_conn)
        while CSR.args.perm > 0:
            CSR.logger.log(
                'CSR iteration %d finished for scenario \'%s\' on ALL databases! Sleep on %d hours' % (
                    iter_num, CSR.args.reindex_scenario, db_name, CSR.args.perm
                ),
                "Info"
            )
            # =====================================================================================
            time.sleep(CSR.args.perm * 3600)
            CSR.logger.log('Ban list len = %d, iteration = %d' % (len(CSR.ban_idxs), iter_num), "Info", do_print=True)
            for v in CSR.ban_idxs:
                CSR.logger.log(str(v), "Info", do_print=True)
            if CSR.sys_conf.ban_hold_iterations % iter_num == 0 and iter_num != 1:
                CSR.ban_idxs = []
                CSR.logger.log('Ban list has been truncated!', "Info")
            for db_name, str_conn in CSR.sys_conf.dbs_dict.items():
                run_on_db(db_name, str_conn)
            iter_num += 1
            # =====================================================================================
    else:
        if CSR.args.db_name not in CSR.sys_conf.dbs_dict:
            CSR.logger.log('Database "%s" not found in config file!' % CSR.args.db_name, "Error", do_print=True)
        else:
            run_on_db(CSR.args.db_name, CSR.sys_conf.dbs_dict[CSR.args.db_name])
            while CSR.args.perm > 0:
                CSR.logger.log(
                    'CSR iteration %d finished for scenario \'%s\' on \'%s\' databases! Sleep on %d hours' % (
                        iter_num, CSR.args.reindex_scenario, CSR.args.db_name, CSR.args.perm
                    ),
                    "Info"
                )
                # =====================================================================================
                time.sleep(CSR.args.perm * 3600)
                CSR.logger.log(
                    'Ban list len = %d, iteration = %d' % (len(CSR.ban_idxs), iter_num),
                    "Info",
                    do_print=True
                )
                for v in CSR.ban_idxs:
                    CSR.logger.log(str(v), "Info", do_print=True)
                if CSR.sys_conf.ban_hold_iterations % iter_num == 0 and iter_num != 1:
                    CSR.ban_idxs = []
                    CSR.logger.log('Ban list has been truncated!', "Info")
                run_on_db(CSR.args.db_name, CSR.sys_conf.dbs_dict[CSR.args.db_name])
                iter_num += 1
                # =====================================================================================

    PSCLogger.instance().stop()
