-- btree index stats query
-- estimates bloat for btree indexes
WITH btree_index_atts AS (
	SELECT
	    nspname,
	    ci.relnamespace,
        ci.relname as index_name, 
        ci.reltuples, 
        ci.relpages, 
        i.indrelid, i.indexrelid,
        ci.relam,
        ct.relname as tablename,
        regexp_split_to_table(indkey::text, ' ')::smallint AS attnum
    FROM pg_index i
    JOIN pg_class AS ci ON i.indexrelid = ci.oid and ci.relkind = 'i'
    JOIN pg_class AS ct ON i.indrelid = ct.oid and ct.relkind in ('r', 'm', 't')
	-- where relkind: r = ordinary table, i = index, S = sequence, v = view, m = materialized view, 
		-- c = composite type, t = TOAST table, f = foreign table
    JOIN pg_namespace n ON n.oid = ci.relnamespace
    JOIN pg_am a ON ci.relam = a.oid
    WHERE a.amname = 'btree'
         AND ci.relpages > 1			-- disable for debug
         AND nspname NOT IN ('pg_catalog', 'information_schema')
		 AND ct.reltuples > 5000		-- disable for debug
),
index_item_sizes AS (
    SELECT
		ind_atts.nspname, ind_atts.relnamespace, ind_atts.index_name, 
		ind_atts.reltuples, ind_atts.relpages, ind_atts.relam, ind_atts.indrelid, ind_atts.indexrelid,
		current_setting('block_size')::numeric AS bs,
		8 AS maxalign,
		24 AS pagehdr,
		CASE WHEN max(coalesce(pg_stats.null_frac, 0)) = 0
			THEN 2
			ELSE 6
		END AS index_tuple_hdr,
		sum((1-coalesce(pg_stats.null_frac, 0)) * coalesce(pg_stats.avg_width, 1024)) AS nulldatawidth
    FROM pg_attribute
    JOIN btree_index_atts AS ind_atts ON pg_attribute.attrelid = ind_atts.indexrelid AND pg_attribute.attnum = ind_atts.attnum
    JOIN pg_stats ON pg_stats.schemaname = ind_atts.nspname
          -- stats for regular index columns
          AND (
			(
			   pg_stats.tablename = ind_atts.tablename AND
			   pg_stats.attname = pg_catalog.pg_get_indexdef(pg_attribute.attrelid, pg_attribute.attnum, TRUE)
			)
          -- stats for functional indexes
          OR   (pg_stats.tablename = ind_atts.index_name AND pg_stats.attname = pg_attribute.attname))
    WHERE pg_attribute.attnum > 0
    GROUP BY 1, 2, 3, 4, 5, 6, 7, 8, 9
),
index_aligned_est AS (
    SELECT maxalign, bs, nspname, index_item_sizes.relnamespace, index_name, reltuples,
        relpages, relam, indrelid, indexrelid,
        coalesce (
            ceil (
                reltuples * ( 6 
                    + maxalign 
                    - CASE
                        WHEN index_tuple_hdr%maxalign = 0 THEN maxalign
                        ELSE index_tuple_hdr%maxalign
                      END
                    + nulldatawidth 
                    + maxalign 
                    - CASE /* Add padding to the data to align on MAXALIGN */
                        WHEN nulldatawidth::integer%maxalign = 0 THEN maxalign
                        ELSE nulldatawidth::integer%maxalign
                      END
                )::numeric 
              / ( bs - pagehdr::NUMERIC )
              +1 )
         , 0 )
      as expected
    FROM index_item_sizes
),
raw_bloat AS (
    SELECT nspname, index_aligned_est.relnamespace, pg_class.relname AS table_name, index_name,
        bs*(index_aligned_est.relpages)::bigint AS totalbytes, expected,
        CASE
            WHEN index_aligned_est.relpages <= expected 
                THEN 0
                ELSE bs*(index_aligned_est.relpages-expected)::bigint 
            END AS wastedbytes,
        CASE
            WHEN index_aligned_est.relpages <= expected
                THEN 0 
                ELSE bs*(index_aligned_est.relpages-expected)::bigint * 100 / (bs*(index_aligned_est.relpages)::bigint) 
            END AS realbloat,
        stat.idx_scan as index_scans, indrelid, index_aligned_est.indexrelid
    FROM index_aligned_est
    JOIN pg_class ON pg_class.oid=index_aligned_est.indrelid
    JOIN pg_stat_user_indexes AS stat ON index_aligned_est.indexrelid = stat.indexrelid
)
-- FldsEnumBasicList
SELECT
	n.oid,
	raw_bloat.nspname,
	table_name,
	index_name,
	(SELECT pg_get_indexdef(raw_bloat.indexrelid)) as def,
	raw_bloat.indexrelid,
	raw_bloat.indrelid,
	exists(select 1 from pg_constraint cn where cn.connamespace = n.oid and cn.conindid = indexrelid) as complex
FROM raw_bloat
JOIN pg_namespace n ON n.oid = raw_bloat.relnamespace
WHERE
    table_name <> 'client'
    AND index_name not like '%_tmp_csr'
    AND index_name not in (
        'session_pool_phone_id_date_c_date_end_x',
        'session_pool_pool_id_end_date_x',
        'ct_keyword_position_value_keyword_position_id_position_date_idx'
    )
    AND realbloat > 40												-- disable for debug
    AND round(wastedbytes/(1024^2)::NUMERIC) > 5 -- "bloat > 5MB"	-- disable for debug
order by wastedbytes desc
limit 100000